/* program to genrerate root-raised-cosine filters for the
 * PFIR
 *
 * formula is:
 *
 * tap(k) = [32768/((pi(1-A)+4A)][1/(1-(4AkR)**2)]*
 *		[sin(kRpi(1-A))/kR + 4Acos(kRpi(1+A))]
 *
 * where:
 *	R is ratio of Baud rate to sample rate
 *	    for 2x oversampling R is 0.5, for 4x it is 0.25
 *
 *	A is excess bandwidth, typically 0.35
 *
 *	k ranges from -31 to +31
 *
 *	The following code has been modified to output the first
 *	32 taps from k=-31 to 0 as 0 to 31 which is what the GC4114 
 *	wants.
 *
 *	Outputs the list of tap values
 */

#include <stdio.h>
#include <math.h>


double ratio=0.5;
double alpha=0.35;

usage(s)
char *s;
{
    fprintf(stderr,"%s\n",s);
    fprintf(stderr,"Usage: rrc\n\tWill ask for alpha in range of 0.0 to 1.0\n");
    fprintf(stderr,"\tWill ask for oversampling ratio\n");
    fprintf(stderr,"\tOversampling ratio is typically 2, 4 or 8 but can be arbitrary\n");
    exit(1);
}

main()
{
    int k;
    double t,a,tg,s,c;
    double gain;
    int taps[32];
    fprintf(stderr,"Enter alpha (0<alpha<1.0): ");
    scanf("%lf",&alpha);
    if (alpha <= 0.0 || alpha >= 1.0) usage("Error: alpha out of range");

    fprintf(stderr,"Enter oversampling ratio (typically 4): ");
    scanf("%lf",&ratio);
    if (ratio <= 1.0)
	usage( "Error: ratio rate must be > 1.0");

    ratio = 1.0/ratio;

    gain = 32767.0/(M_PI*(1.0-alpha)+4.0*alpha);

    /*
    fprintf(stderr,"ratio=%f, gain=%f\n",ratio,gain);
    */
    for (k=0;k<31;k++) {
	t = 4.0*alpha*ratio*((double) (31-k));
	tg = gain/(1.0-(t*t));
	a = ratio*((double) (31-k));
	s =  sinpi((1.0-alpha)*a);
	c =  cospi((1.0+alpha)*a);
	t = tg * ( s/(ratio*((double) (31-k))) + 4.0*alpha*c);
	taps[k] = irint(t);
	printf("%d\n",taps[k]);
    }
    taps[31]=32767.0;
    printf("32767\n");
    for (k=30;k>=0;k--) printf("%d\n",taps[k]);
}
