/* command file generator for gc4016 */

#include <stdio.h>
#include <string.h>

typedef struct {
    int add_to;
    int blankcnt;
    int blanken;
    int blanksync;
    int cex;		/* constant exponent */
    int cicscale;
    int coarse;		/* shift after CIC */
    int deccount;
    int deccountsync;
    int dither;
    int dithersync;
    int enable;
    int finegain;
    int flushsync;
    int freq;
    int freqsync;
    int gainsync;
    int idlyqtr;
    int idlyhlf;
    int ifa;		/* input format A/B select */
    int ifmt;		/* input format */
    int in;		/* input selection */
    int ionly;
    int msbpol;
    int ncosync;
    int negctl;
    int nosym21;
    int nosym63;
    int ofselect;
    int out20;
    int outchan;
    int peaksync;
    int peakthresh;
    int peakselect;
    int phase;
    int phasesync;
    int qdlyqtr;
    int qdlyhlf;
    int qonly;
    int real2x;
    int scex;
    int test;
    double coef21[21];
    double coef63[63];
} CHAN;

typedef struct {
    int resampratio;
    int filterselect;
} OCHAN;

/* a structure to capture the various general control bits 
 * registers 0-7 and the output control bits (page 98) 
 */

typedef struct {
    int fourbit;
    int blocklength;
    int bnperword;
    int ck2xen;
    int ck2xtest;
    int ckdiv;
    int cklossdetect;
    int cnt;
    int countersync;
    int counttest;
    int diagsource;
    int diagsync;
    int encko;
    int enpar;
    int enrdy;
    int ensa;
    int ensb;
    int ensc;
    int ensd;
    int ensfs;
    int enso;
    int en4fs;
    int fl;
    int globalen;
    int invcko;
    int invrdy;
    int invsfs;
    int latewrite;
    int link;
    int lvds;
    int master;
    int nibble;
    int ni;
    int no;
    int oneshot;
    int osmode;
    int os;
    int osync;
    int outen;
    int outputsync;
    int paden;
    int parallel;
    int rdywidth;
    int resen;
    int rev;
    int sfc;
    int ssa;
    int ssb;
    int ssc;
    int ssd;
    int tagai;
    int tagaq;
    int tagbi;
    int tagbq;
    int tagci;
    int tagcq;
    int tagdi;
    int tagdq;
    int tagen;
    int tag22;
    int wordsperframe;
} GEN;

#define NF 4
FILE *fcmd;
FILE *ftbl;
FILE *fout;
char name[128];
CHAN chan[4];
OCHAN ochan[4];
double rescoef[512];
int nrm, nochan, nosymres, nrestaps, resclkdiv;
int ressync, ratiosync, ratiomap;
int realout, round, sfiltcon, finalshift;

#define MBIG 0x80000000
#define  BIG 0x7fffffff

#define TI 1
#define GC100 2
#define SIM 3
#define INI 4

int page;
int err;
char *seps;
double rescf[4][512];
int nresf;
char line[1000];

GEN gen;
int pdoc, outstyle;

  FILE *
myopen(firstname,lastname)
    char *firstname, *lastname;
    {
    FILE *f;
    char line[1000];

    sprintf(line,"%s.%s", firstname, lastname);
    fprintf(stderr,"Opening output file: %s\n\n",line);
    f = fopen(line, "w");
    if(f == NULL)  {
	fprintf(stderr, "Couldn't open %s\n", line);
	exit(1);
    }
    return(f);
}

main(argc, argv) 
  int argc; char **argv;
     {
     int i, j, time, ntime;
     int readit;

     outstyle=0;
     if(argc==1) {
	fprintf(stderr, "Usage: cmd4016 [-TI | -GC100 | -SIM] controlfile\n");
	exit(2);
     }
     if (argc==2) {
        strcpy(name,argv[1]);
     } else if (argc==3) {
	if (*argv[1] == '-') {
	    switch(*(argv[1]+1)) {
	    case 'T': outstyle = TI;break;
	    case 't': outstyle = TI;break;
	    case 'g': outstyle = GC100;break;
	    case 'G': outstyle = GC100;break;
	    case 's': outstyle = SIM;break;
	    case 'S': outstyle = SIM;break;
	    case 'i': outstyle = INI;break;
	    case 'I': outstyle = INI;break;
	    default:
		fprintf(stderr, "Unrecognized print option %s\n", argv[1]);
		fprintf(stderr, "Options are: TI, gc100, sim\n");
		exit(2);
	    }
	}
        strcpy(name,argv[2]);
     } else {
	fprintf(stderr, "Usage: cmd4016 [-TI | -GC100 | -SIM] controlfile\n");
	exit(2);
     }

    ftbl = myopen(name,"tbl");
     cmdin(name);
     fclose(fout);
     fclose(ftbl);

}


/* command parser for sim4016 */
/* general format :
  variable_name integer_value
  cfir_coef filename
  pfir_coef filename
  res_coef number filename
  channel number
*/

cmdin(cmds) 
    char *cmds;
    {

    int unrec, linecnt, infile;
    int chkochan[4], i, temp, sane;
    char *c;
    CHAN *cpi;

    pdoc = 1;

    if((fcmd = fopen(cmds, "r"))==NULL) {
	fprintf(stderr, "Couldn't open command file %s\n", cmds);
	fprintf(ftbl, "Couldn't open command file %s\n", cmds);
	exit(1);
    }
    seps = " \t,=\r";

    err = 0;
    sane = 1; /* generate lots of output unless requested not to */
    page = -99; /* start off so the first write will set the page */
    nresf = 0;

    for(i=0; i<4; i++) chkochan[i] = 0;

    for(linecnt=1; fgets(line, 1000, fcmd)!= NULL; linecnt++) {
    line[strlen(line)-1] = '\0';
    unrec = 0;
    c = strtok(line, seps);
    if(c)
    switch (*c) {
    case '#':
    case ';':
    case '/': /* a comment line */ break;

    case '4':
        if(getint(c,"4_bit_address",&gen.fourbit,0,1)) break;
	unrec = 1; break;
    case 'a':
	if(getint(c,"ab_sel",&cpi->ifa ,0,1)) break;
	if(getint(c,"add_a_to_b",&chan[0].add_to ,0,1)) break;
	if(getint(c,"add_b_to_c",&chan[1].add_to ,0,1)) break;
	if(getint(c,"add_c_to_d",&chan[2].add_to ,0,1)) break;
	unrec=1; break;
    case 'b':
	if(getint(c,"big_scale",&temp,0,7)) {
	    cpi->cicscale = (cpi->cicscale & ~0x38) | (temp<<3);
	    break;
	}
	if(getint(c,"block_size",&gen.blocklength,0,7)) break;
	if(getint(c,"bits_per_word",&gen.bnperword,0,7)) break;
	unrec = 1; break;
    case 'c':
	if(strcmp(c,"cfir_coef")==0) {readcoef(cpi->coef21,21); break;}
	if(getint(c,"ch_reset",&temp,0,1)) {
		cpi->enable = 1-temp;
		break;
	}
	if(getint(c, "channel",&temp ,0,3)) { cpi = &chan[temp]; break;}
	if(getint(c,"chan_map_a",&chan[0].outchan ,0,3)) break;
	if(getint(c,"chan_map_b",&chan[1].outchan ,0,3)) break;
	if(getint(c,"chan_map_c",&chan[2].outchan ,0,3)) break;
	if(getint(c,"chan_map_d",&chan[3].outchan ,0,3)) break;
        if(getint(c,"ck_2x_en",&gen.ck2xen,0,1)) break;
        if(getint(c,"ck_2x_test",&gen.ck2xtest,0,1)) break;
        if(getint(c,"ck_loss_detect",&gen.cklossdetect,0,1)) break;
        if(getint(c,"cnt",&gen.cnt,0,65535)) break;
	if(getint(c,"coarse",&cpi->coarse ,0,7)) break;
	if(getint(c,"copy_channel",&temp ,0,3)) {
	    copy_channel(cpi,temp);
	    break;
	}
        if(getint(c,"counter_sync",&gen.countersync,0,7)) break;
        if(getint(c,"count_test",&gen.counttest,0,1)) break;
	unrec = 1; break;
    case 'd':
	if(getint(c,"dec", &cpi->deccount,0,4095)) break;
	if(getint(c,"dec_sync",&cpi->deccountsync ,0,7)) break;
        if(getint(c,"diag_source",&gen.diagsource,0,3)) break;
        if(getint(c,"diag_sync",&gen.diagsync,0,7)) break;
	if(getint(c,"dither_sync",&cpi->dithersync ,0,7)) {
	    if(cpi->dithersync>5) cpi->dither = 0;
	    else cpi->dither= 1;
	    break;
	}
	unrec = 1; break;
    case 'e':
        if(getint(c,"edge_write",&gen.latewrite,0,1)) break;
        if(getint(c,"en_p0",&gen.ensa,0,1)) break;
        if(getint(c,"en_p1",&gen.ensb,0,1)) break;
        if(getint(c,"en_p2",&gen.ensc,0,1)) break;
        if(getint(c,"en_p3",&gen.ensd,0,1)) break;
        if(getint(c,"en_par",&gen.enpar,0,1)) break;
        if(getint(c,"en_rdy",&gen.enrdy,0,1)) break;
        if(getint(c,"en_sck",&gen.encko,0,1)) break;
        if(getint(c,"en_sfs",&gen.ensfs,0,1)) break;
        if(getint(c,"en_so",&gen.enso,0,1)) break;
        if(getint(c,"en_4_fs",&gen.en4fs,0,1)) break;
	unrec = 1; break;
    case 'f':
	if(getint(c,"final_shift", &finalshift,0,15)) break;
	if(getint(c,"filter_sel_0",&ochan[0].filterselect ,0,3)) break;
	if(getint(c,"filter_sel_1",&ochan[1].filterselect ,0,3)) break;
	if(getint(c,"filter_sel_2",&ochan[2].filterselect ,0,3)) break;
	if(getint(c,"filter_sel_3",&ochan[3].filterselect ,0,3)) break;
	if(getint(c,"fine_gain",&cpi->finegain ,0,16383)) break;
        if(getint(c,"frame_length",&gen.fl,0,63)) break;
	if(getint(c,"flush_sync",&cpi->flushsync ,0,7)) break;
	if(getint(c,"freq",&cpi->freq,-BIG,BIG)) break;
	if(getint(c,"freq_sync",&cpi->freqsync ,0,7)) break;
	unrec = 1; break;
    case 'g':
	if(getint(c,"gain_sync",&cpi->gainsync ,0,7)) break;
        if(getint(c,"global_reset",&gen.globalen,0,1)) break;
    case 'i':
	if(getint(c,"idly_cfir",&cpi->idlyqtr ,0,1)) break;
	if(getint(c,"idly_pfir",&cpi->idlyhlf ,0,1)) break;
	if(getint(c,"in_fmt",&cpi->ifmt ,0,3)) break;
	if(getint(c,"input_sel",&cpi->in ,0,7)) break;
        if(getint(c,"inv_rdy",&gen.invrdy,0,1)) break;
        if(getint(c,"inv_sck",&gen.invcko,0,1)) break;
        if(getint(c,"inv_sfs",&gen.invsfs,0,1)) break;
	if(getint(c,"ionly",&cpi->ionly ,0,1)) break;
	unrec = 1; break;
    case 'l':
        if(getint(c,"link",&gen.link,0,1)) break;
        if(getint(c,"lvds",&gen.lvds,0,1)) break;
	unrec = 1; break;
    case 'm':
        if(getint(c,"master",&gen.master,0,1)) break;
	if(getint(c,"mix20b",&cpi->out20 ,0,1)) break;
	if(getint(c,"msb_pol",&cpi->msbpol ,0,1)) break;
	unrec = 1; break;
    case 'n':
	if(getint(c,"nc",&nochan ,0,3)) break;
	if(getint(c,"nco_sync",&cpi->ncosync ,0,7)) break;
	if(getint(c,"neg_ctl",&cpi->negctl ,0,15)) break;
	if(getint(c,"nf",&sfiltcon ,0,3)) break;
        if(getint(c,"nibble",&gen.nibble,0,1)) break;
	if(getint(c,"nm",&nrm ,5,62)) break;
	if(getint(c,"no_sym_cfir",&cpi->nosym21 ,0,1)) break;
	if(getint(c,"no_sym_pfir",&cpi->nosym63 ,0,1)) break;
	if(getint(c,"no_sym_res",&nosymres ,0,1)) break;
        if(getint(c,"nserial",&gen.no,0,3)) break;
	if(getint(c,"nzeros",&cpi->blankcnt ,0,15)) break;
	unrec = 1; break;
    case 'o':
        if(getint(c,"one_shot",&gen.oneshot,0,1)) break;
        if(getint(c,"os_mode",&gen.osmode,0,1)) break;
        if(getint(c,"out_blk_sync",&gen.osync,0,7)) break;
        if(getint(c,"out_blk_reset",&gen.outen,0,1)) break;
        if(getint(c,"output_mode",&gen.os,0,3)) break;
        if(getint(c,"output_order",&gen.ni,0,3)) break;
        if(getint(c,"output_sync",&gen.outputsync,0,7)) break;
	unrec = 1; break;
    case 'p':
        if(getint(c,"pad_reset",&gen.paden,0,1)) break;
        if(getint(c,"parallel",&gen.parallel,0,1)) break;
	if(strcmp(c,"pfir_coef")==0) {readcoef(cpi->coef63,63); break;}
	if(getint(c,"phase",&cpi->phase ,0,65535)) break;
	if(getint(c,"phase_sync",&cpi->phasesync ,0,7)) break;
	if(getint(c,"peak_sync",&cpi->peaksync ,0,7)) break;
	if(getint(c,"peak_select",&cpi->ofselect,0,3)) break;
	if(getint(c,"peak_mode",&cpi->peakselect ,0,1)) break;
	if(getint(c,"peak_threshold",&cpi->peakthresh ,0,3)) break;
	if(strcmp(c, "print") == 0) {
	    c=strtok(NULL,seps);
	    if (outstyle) {
	       fprintf(stderr,"Warning: command \"print %s\" ignored, overridden on command line\n",c);
	       fprintf(ftbl,"Warning: command \"print %s\" ignored, overridden on command line\n",c);
	       break;
	    }
	    if     (*c == 'T') outstyle = TI;
	    else if(*c == 't') outstyle = TI;
	    else if(*c == 'g') outstyle = GC100;
	    else if(*c == 'G') outstyle = GC100;
	    else if(*c == 's') outstyle = SIM;
	    else if(*c == 'S') outstyle = SIM;
	    else if(*c == 'i') outstyle = INI;
	    else if(*c == 'I') outstyle = INI;
	    else {
		fprintf(stderr, "Unrecognized print command %s\n", c);
		fprintf(stderr, "Options are: TI, gc100, sim\n");
		fprintf(ftbl, "Unrecognized print command %s\n", c);
		fprintf(ftbl, "Options are: TI, gc100, sim\n");
		unrec = 1;
	    }
	    break;
	}
	unrec = 1; break;
    case 'q':
	if(getint(c,"qdly_cfir",&cpi->qdlyqtr ,0,1)) break;
	if(getint(c,"qdly_pfir",&cpi->qdlyhlf ,0,1)) break;
	if(getint(c,"qonly",&cpi->qonly ,0,1)) break;
	unrec = 1; break;
    case 'r':
	if(getint(c,"ratio_0",&ochan[0].resampratio ,MBIG,BIG)) break;
	if(getint(c,"ratio_1",&ochan[1].resampratio ,MBIG,BIG)) break;
	if(getint(c,"ratio_2",&ochan[2].resampratio ,MBIG,BIG)) break;
	if(getint(c,"ratio_3",&ochan[3].resampratio ,MBIG,BIG)) break;
        if(getint(c,"rdy_width",&gen.rdywidth,0,1)) break;
        if(getint(c,"real_only",&realout,0,1)) break;
	if(getint(c,"res_clk_div",&resclkdiv,0,255)) break;
	if(getint(c,"res_coef",&temp,0,3)) {
	    nresf |= (1<<temp); /* mark as having read in coefs for later use */
	    readcoef(rescf[temp],512); break;}
        if(getint(c,"resampler_reset",&gen.resen,0,1)) break;
        if(getint(c,"reverse_iq",&gen.rev,0,1)) break;
	if(getint(c,"round",&round ,0,3)) break;
	if(getint(c,"res_sync",&ressync ,0,7)) break;
	if(getint(c,"ratio_sync",&ratiosync ,0,7)) break;
	if(getint(c,"ratio_map",&ratiomap ,0,255)) break;
	unrec = 1; break;
    case 's':
	if(getint(c,"scale",&temp,0,5)) {
	    cpi->cicscale = (cpi->cicscale & ~0x7) | temp;
	    break;
	}
        if(getint(c,"sck_rate",&gen.ckdiv,0,15)) break;
        if(getint(c,"sfs_mode",&gen.sfc,0,3)) break;
	if(getint(c,"shift",&cpi->cex ,0,7)) break;
	if(getint(c,"splitiq",&cpi->real2x ,0,1)) break;
        if(getint(c,"smux_0",&gen.ssa,0,3)) break;
        if(getint(c,"smux_1",&gen.ssb,0,3)) break;
        if(getint(c,"smux_2",&gen.ssc,0,3)) break;
        if(getint(c,"smux_3",&gen.ssd,0,3)) break;
	unrec = 1; break;
    case 't':
        if(getint(c,"tag_ai",&gen.tagai,0,15)) break;
        if(getint(c,"tag_aq",&gen.tagaq,0,15)) break;
        if(getint(c,"tag_bi",&gen.tagbi,0,15)) break;
        if(getint(c,"tag_bq",&gen.tagbq,0,15)) break;
        if(getint(c,"tag_ci",&gen.tagci,0,15)) break;
        if(getint(c,"tag_cq",&gen.tagcq,0,15)) break;
        if(getint(c,"tag_di",&gen.tagdi,0,15)) break;
        if(getint(c,"tag_dq",&gen.tagdq,0,15)) break;
        if(getint(c,"tag_en",&gen.tagen,0,1)) break;
        if(getint(c,"tag_22",&gen.tag22,0,1)) break;
	if(getint(c,"test",&cpi->test ,0,1)) break;
	unrec = 1; break;
    case 'u':
	if(getint(c,"use_shift",&cpi->scex ,0,1)) break;
	unrec = 1; break;
    case 'w':
        if(getint(c,"words_per_frame",&gen.wordsperframe,0,7)) break;
	unrec = 1; break;
    case 'z':
	if(getint(c,"zpad_en",&cpi->blanken ,0,1)) break;
	if(getint(c,"zpad_sync",&cpi->blanksync ,0,7)) break;
	unrec = 1; break;
    default:
	unrec = 1; break;
    }
    if(unrec) {
	err = 1;
	fprintf(stderr, "Unrecognized command on line %d:%s\n", linecnt, line);
	fprintf(ftbl, "Unrecognized command on line %d:%s\n", linecnt, line);
    }}

    fclose(fcmd);

    /* now open output files, *.tbl is always generated */

    switch(outstyle) {
    case TI:
	fout = myopen(name,"ti");
        if (gen.fourbit == 0) {
	    fprintf (stderr,"Warning: 4_bit_address is being forced on for the TI mode\n");
	    fprintf (ftbl,"Warning: 4_bit_address is being forced on for the TI mode\n");
	}
	gen.fourbit=1;
	break;
    case GC100:
	fout = myopen(name,"gc100");
        if (gen.fourbit == 1) {
	    fprintf (stderr,"Warning: 4_bit_address is being forced off for the GC100 mode\n");
	    fprintf (ftbl,"Warning: 4_bit_address is being forced off for the GC100 mode\n");
	}
	gen.fourbit=0;
	break;
    case INI:
	fout = myopen(name,"ini");
        if (gen.fourbit == 1) {
	    fprintf (stderr,"Warning: 4_bit_address is being forced off for the INI mode\n");
	    fprintf (ftbl,"Warning: 4_bit_address is being forced off for the INI mode\n");
	}
	gen.fourbit=0;
	break;
    case SIM:
	fout = myopen(name,"sim");
        if (gen.fourbit == 1) {
	    fprintf (stderr,"Warning: 4_bit_address is being forced off for the SIM mode\n");
	    fprintf (ftbl,"Warning: 4_bit_address is being forced off for the SIM mode\n");
	}
	gen.fourbit=0;
	break;
    default: break;
    }

    for(i=0; i<4; i++) {
      if(chan[i].enable) {
         checkcoef(sprintf(line, "cfir_%d", i), chan[i].coef21, 21, 16,
	      1-chan[i].nosym21, 1);
         checkcoef(sprintf(line, "pfir_%d", i), chan[i].coef63, 63, 16,
	      1-chan[i].nosym63, 1);
	}
    }
    checkrescoef();
    printgen();
    for(i=0; i<4; i++) writechan(i);
    printres();
    printout();
    if(err) exit(1);
    /* map values to make checking easier */
    if(gen.ni==2) gen.ni=3;
    if(gen.no==2) gen.no=3;
    if(gen.blocklength==2) gen.blocklength=3;
    sanity(sane, &gen);
}

getint(c,str, var, vmin, vmax) 	
    char *c, *str;
    int *var, vmin, vmax;
    {
    char *cc;
    int d;

    if(strcmp(c,str)==0) {
        cc = strtok(NULL, seps);
	if (strncmp(cc,"0x",2)==0) sscanf(cc,"0x%x",&d);
        else sscanf(cc, "%d", &d);
	if((d < vmin) || (d > vmax)) {
	    fprintf(stderr, "%s must be between %d and %d\n", str, vmin, vmax);
	    fprintf(ftbl, "%s must be between %d and %d\n", str, vmin, vmax);
	    err = 1;
	}
	*var = d;
	return(1);
    }
    return(0);
}

#define COPY_CHAN(x) new->x = old->x

copy_channel(new,number)
    CHAN *new;
    int number;
{
    /* copy configuration from chan[number] to new */
    int i;
    CHAN *old;
    old = &chan[number];
    COPY_CHAN(blankcnt);
    COPY_CHAN(blanken);
    COPY_CHAN(blanksync);
    COPY_CHAN(cex);
    COPY_CHAN(cicscale);
    COPY_CHAN(coarse);
    COPY_CHAN(deccount);
    COPY_CHAN(deccountsync);
    COPY_CHAN(dither);
    COPY_CHAN(dithersync);
    COPY_CHAN(enable);
    COPY_CHAN(finegain);
    COPY_CHAN(flushsync);
    COPY_CHAN(freq);
    COPY_CHAN(freqsync);
    COPY_CHAN(gainsync);
    COPY_CHAN(idlyqtr);
    COPY_CHAN(idlyhlf);
    COPY_CHAN(ifa);
    COPY_CHAN(ifmt);
    COPY_CHAN(in);
    COPY_CHAN(ionly);
    COPY_CHAN(msbpol);
    COPY_CHAN(ncosync);
    COPY_CHAN(negctl);
    COPY_CHAN(nosym21);
    COPY_CHAN(nosym63);
    COPY_CHAN(ofselect);
    COPY_CHAN(out20);
    COPY_CHAN(outchan);
    COPY_CHAN(peaksync);
    COPY_CHAN(peakthresh);
    COPY_CHAN(peakselect);
    COPY_CHAN(phase);
    COPY_CHAN(phasesync);
    COPY_CHAN(qdlyqtr);
    COPY_CHAN(qdlyhlf);
    COPY_CHAN(qonly);
    COPY_CHAN(real2x);
    COPY_CHAN(scex);
    COPY_CHAN(test);
    for(i=0;i<63;i++) new->coef63[i] = old->coef63[i];
    for(i=0;i<21;i++) new->coef21[i] = old->coef21[i];
}

readcoef(taps, maxtaps)
    double taps[];
    int maxtaps;
    {
    char *c;
    FILE *fp;
    int i, temp;

    c = strtok(NULL, seps);
    if((fp = fopen(c, "r")) == NULL) {
	fprintf(stderr, "Couldn't open coefficient file: %s\n", c);
	fprintf(ftbl, "Couldn't open coefficient file: %s\n", c);
	err = 1;
	exit(1);
	return(0);
    }

    for(i=0; i<maxtaps; i++) taps[i]=0;
    for(i=0; i<maxtaps; i++) {
	if(fscanf(fp, "%lf", &taps[i]) != 1) break;
    }
    if((i==maxtaps) && (fscanf(fp, "%lf", &temp) == 1)) {
	fprintf(stderr, "Too many coefficients in file: %s\n", c);
	fprintf(ftbl, "Too many coefficients in file: %s\n", c);
	err = 1;
    }
    fclose(fp);
    return(i);
    }

checkcoef(name, cf, n, bits, sym, odd)
  char *name;
  int n, bits, sym, odd;
  double cf[];
  {
  double coef[512];
  int m, i, j, ntaps, i0;
    m = 1<<(bits-1);
    for(i=0; i<n; i++) {
	if((cf[i]< -m)||(cf[i]>=m)) {
	  fprintf(stderr, "Coefficients for %s must be %d bits\n", name, bits);
	  fprintf(ftbl, "Coefficients for %s must be %d bits\n", name, bits);
	  if(cf[i]==-m) {
	    fprintf(stderr, "Coefficients for %s can not be %d\n", name, -m);
	    fprintf(ftbl, "Coefficients for %s can not be %d\n", name, -m);
	  }
	  err = 1;
	  break;
        }
    }
    for(i0=0; i0<n/2; i0++) if(cf[i0]) break;
    for(i=n-1; i>-1; i--) if(cf[i]) break;
    ntaps = i-i0+1;

    if(sym==0) { /* no symmetry */
    if(ntaps>(n+odd)/2) {
       fprintf(stderr, 
       "In nonsymmetric mode the maximum number of taps for %s is %d\n", 
       name, (n+odd)/2);
       fprintf(ftbl, 
       "In nonsymmetric mode the maximum number of taps for %s is %d\n", 
       name, (n+odd)/2);
       err = 1;
    }} else { /* symmetry */
    if(ntaps > n) {
       fprintf(stderr, 
       "In symmetric mode the maximum number of taps for %s is %d\n", name, n);
       fprintf(ftbl, 
       "In symmetric mode the maximum number of taps for %s is %d\n", name, n);
       err = 1;
    }
    if(odd) { /* check for odd symmetry */
    /* NOTE odd also happens to imply cfir or pfir so move coefs to
       the proper place if needed. */
       if((ntaps&1)==0) {
	  fprintf(stderr, "Filter %s must be odd length if symmetric\n", name);
	  fprintf(ftbl, "Filter %s must be odd length if symmetric\n", name);
	  err = 1;
	}
	for(j=0; j<ntaps/2; j++) {
	   if(cf[i0+j]!=cf[i0+ntaps-j-1]) {
	  fprintf(stderr, "Filter %s must be odd symmetric\n", name);
	  fprintf(ftbl, "Filter %s must be odd symmetric\n", name);
	  err = 1;
	  break;
	}
	if(err==0) { /* now put the coefs in their proper place */
	for(j=0,i=i0; j<ntaps; i++,j++) coef[j] = cf[i];
	for(j=0; j<n; j++) cf[j] = 0;
	for(j=0, i=(n-ntaps)/2; j<ntaps; j++,i++) cf[i] = coef[j];
	}
    }
    } else { /* even symmetry */
     /* check for even symmetry */
       if((ntaps&1)==1) {
	  fprintf(stderr, "Filter %s must be even length if symmetric\n", name);
	  fprintf(ftbl, "Filter %s must be even length if symmetric\n", name);
	  err = 1;
	}
	for(j=0; j<ntaps/2; j++) {
	   if(cf[i0+j]!=cf[i0+ntaps-j-1]) {
	  fprintf(stderr, "Filter %s must be even symmetric\n", name);
	  fprintf(ftbl, "Filter %s must be even symmetric\n", name);
	  err = 1;
	  break;
	}
	if(err==0) { /* now put the coefs in their proper place */
	for(j=0,i=i0; j<ntaps; i++,j++) coef[j] = cf[i];
	for(j=0; j<n; j++) cf[j] = 0;
	for(j=0, i=(n-ntaps)/2; j<ntaps; j++,i++) cf[i] = coef[j];
	}
    }
    }
    }
    }

int ndelay;

checkrescoef() {
     /* resampler coefficient checks
	 12 bit coefficients
	 not too many of them
	 symmetry if turned on
	 warnings if we did not read in all appro filters
	 error if we use filter beyond number specified
	 center properly symmetric filters
	 interleave multiple filters
     nm, sfiltcon
     */
	 
     int nf, temp, i, j , qtaps, k;


     nf  = sfiltcon;
     temp = 256*(2-nosymres)/((nf+1)*(nrm+1));
     for(ndelay=4; ndelay<=temp; ndelay = ndelay*2) ;
     ndelay /= 2;
     qtaps = (nrm+1)*ndelay;
     nrestaps = qtaps;

     /* check for coefficient symmetry, size, count */
     for(i=0; i<=nf; i++) {
     if(nf>0) sprintf(line, "resampler coefficients filter %d", i);
     else sprintf(line, "resampler coefficients");
     for(k=0; k<512; k++) if(rescf[i][k] != 0) break;
     for(j=512; j>0; j--) if(rescf[i][j] != 0) break;
     if((j-k+1) > nrestaps) {
	 fprintf(stderr, "%s %d taps\n",
	 "Each resampler filter may have no more than", nrestaps);
	 fprintf(ftbl, "%s %d taps\n",
	 "Each resampler filter may have no more than", nrestaps);
	 err = 1;
     }
     checkcoef(line, rescf[i], qtaps*(1+nosymres), 12, 1-nosymres, 0);
     }

     /* check that each active channel selects a filter that has been
      *	read in and is less than the number filters defined
      */
    for(i=0; i<nochan; i++) if(ochan[i].filterselect > nf) {
	fprintf(stderr, "In channel %d you requested a resampler %s",
	   i, "filter beyond those specified.\n");
	fprintf(ftbl, "In channel %d you requested a resampler %s",
	   i, "filter beyond those specified.\n");
	err=1;
    }

    /* check that we read in filters for all specified */
    for(i=0; i<=nf; i++) if((nresf & (1<<i))==0) {
	fprintf(stderr, "Warning. Resampler filter %d not read in.\n", i);
	fprintf(ftbl, "Warning. Resampler filter %d not read in.\n", i);
    }

    /* check for reading in filters beyond those specified */
    for(i=nf+1; i<4; i++) if((nresf & (1<<i))!=0) {
	fprintf(stderr, "Warning. Resampler filter %d read in %s\n",
	i, nf, "but nf is only %d.");
	fprintf(ftbl, "Warning. Resampler filter %d read in %s\n",
	i, nf, "but nf is only %d.");
    }

    /* now interleave filters */
    for(j=0; j<=nf; j++) 
	for(i=0; i<qtaps; i++) rescoef[i*(nf+1)+j]=rescf[j][i];
}


writerescoef(i)
    int i;
    {
    int j, k, rc;
    for(k=0; k<4; k++) {
	j = i+k;
	rc = rescoef[j];
        writec(32+j/8, 16+2*(j&0x7), rc    &0xff);
        writec(32+j/8, 17+2*(j&0x7),(rc>>8)&0xff);
    }
    }

printgen() {	/* program general and output page registers */
    int j, c;
    int r[32];
    /* general registers */

    for(j=0; j<32; j++) r[j]=0;
    /* general mode */
    r[0]= c = gen.cklossdetect + 2*gen.ck2xtest + (gen.ck2xen<<2) + 
	(gen.latewrite<<3) + (gen.resen<<4) + (gen.paden<<5)+
	(gen.outen<<6) + (gen.globalen<<7);
    writec(0,0,c);
    /* write 0 to status and page, skip checksum */
    writec(0,1,0);
    writec(0,2,0);

    /* general sync */
    r[4] = c = gen.diagsync + (gen.outputsync<<3) + (gen.fourbit<<6) + (gen.lvds<<7);
    writec(0,4,c);

    /* counter sync */
    r[5] = c = gen.diagsource + (gen.countersync<<2) + (gen.counttest<<5) +
    (gen.osmode<<6) + (gen.oneshot<<7);
    writec(0,5,c);

    /* counter */
    writec(0,6,(r[6]=(gen.cnt&0xff)));
    writec(0,7,(r[7]=((gen.cnt>>8)&0xff)));
    if (outstyle==INI) writeini(fout,-1,r);
    if(pdoc) {
        fprintf(ftbl, "Globals:\n  ");
        for(j=0; j<8; j++) fprintf(ftbl, " %2x",r[j]);
        fprintf(ftbl, "\n");
    }
    if(outstyle==GC100) {
        fprintf(fout, "# ------ Globals ------\nwrite_block 0xd000 \\\n");
        for(j=0; j<8; j++) fprintf(fout, " 0x%02x",r[j]);
        fprintf(fout, "\n");
    }
}


printout() {	/* program output page registers */
    int j, c;
    int r[32];

    for(j=0; j<32; j++) r[j]=0;
    /* output registers (page 98) */
    /* tristate register */
    r[16]=c = gen.encko + (gen.enrdy<<1) + (gen.ensfs<<2) + (gen.ensa<<3) +
	(gen.ensb<<4) + (gen.ensc<<5) + (gen.ensd<<6) + (gen.enpar<<7);
    writec(98,16,c);

    /* output sync register */
    r[17]=c = gen.invcko + (gen.invrdy<<1) + (gen.invsfs<<2) + (gen.tagen<<3) +
	(gen.rdywidth<<4) + (gen.osync<<5);
    writec(98,17,c);

    /* output format register */
    r[18]=c = gen.link + (gen.nibble<<1) + (gen.parallel<<2) + (gen.master<<3) +
	(realout<<4) + (gen.os<<5) + (gen.rev<<7);
    writec(98,18,c);

    /* output frame length */
    r[19] = c = gen.fl + (gen.sfc<<6);
    writec(98,19, c);

    /* output wordsize register */
    r[20] = c = gen.wordsperframe + (gen.bnperword<<3) + (gen.blocklength<<6);
    writec(98,20,c);

    /* output clock register */
    r[21] = c = gen.ckdiv+(gen.no<<4)+(gen.ni<<6);
    writec(98,21,c);

    /* serial port selection register */
    r[22] = c = gen.ssa + (gen.ssb<<2) + (gen.ssc<<4) + (gen.ssd<<6);
    writec(98,22,c);

    /* tags registers */
    writec(98,23, (r[23]= gen.tagai + (gen.tagaq<<4)));
    writec(98,24, (r[24]= gen.tagbi + (gen.tagbq<<4)));
    writec(98,25, (r[25]= gen.tagci + (gen.tagcq<<4)));
    writec(98,26, (r[26]= gen.tagdi + (gen.tagdq<<4)));
    writec(98,28, (r[28]= gen.en4fs + (gen.enso<<1)));
    r[27]=0;

    
    if (outstyle==INI) writeini(fout,98,r);
    if(pdoc) {
        fprintf(ftbl, "\nOutput Page:\n98");
        for(j=16; j<29; j++) fprintf(ftbl, " %2x",r[j]);
        fprintf(ftbl, "\n");
    }
    if(outstyle==GC100) {
        fprintf(fout, "\n# ------ Output Page ------\n");
	writeit_GC100(fout,2,98*2);
	fprintf(fout,"write_block 0xd010 \\\n");
        for(j=16; j<29; j++) fprintf(fout, " 0x%02x",r[j]);
        fprintf(fout, "\n");
    }


}


writechan(i)  int i;
{		/* programs a channel register/ram */
  /* writec pagenumber address datavalue */
  /* all in decimal */

  int k, r[32];
  int j, c;
  char ch;
  switch(i) {
  case 0: ch = 'A'; break;
  case 1: ch = 'B'; break;
  case 2: ch = 'C'; break;
  case 3: ch = 'D'; break;
  default: break;
  }
  if (pdoc) fprintf(ftbl,"\nChannel %c:\n",ch);
  if (outstyle==GC100) fprintf(fout,"\n# ------ Channel %c ------\n",ch);
  if(chan[i].enable ==0) { /* just write to power down this channel */
  	/* and skip the rest of the writting */
  	writec(7+i*8,16, 0x80);
	if (pdoc) fprintf(ftbl,"Powered down\n%2d %2x\n",7+i*8,0x80);
	if (outstyle==GC100) {
 	    fprintf(ftbl,"# Powered down\n");
	    writeit_GC100(fout,2,2*(7+i*8));
	    writeit_GC100(fout,16,0x80);
	}
	return(0);
    }

  /* pages 0&1  fir1 coefficient storage */
  if (pdoc) fprintf(ftbl,"CFIR Coefficients:");
  if (outstyle==GC100) fprintf(fout,"# CFIR Coefficients:");
  for(j=0; j<11; j++) {
     c = chan[i].coef21[j];
     writec((j/8)+i*8,16+2*(j%8), c & 0xff);
     writec((j/8)+i*8,17+2*(j%8),(c>>8)&0xff);
     if(pdoc) {
         if((j%8)==0) fprintf(ftbl,"\n%2d",j/8+i*8);
         fprintf(ftbl, " %2x %2x",c & 0xff, (c>>8) & 0xff);
     }
    if(outstyle==GC100) {
	if (j%8==0) {
          fprintf(fout, "\n");
	  writeit_GC100(fout,2,2*((j/8)+i*8));
	  fprintf(fout,"write_block 0xd010 \\\n");
	}
        fprintf(fout, " 0x%02x 0x%02x",c & 0xff, (c>>8) & 0xff);
    }
  }


  /* pages 2-5 fir2 coefficient storage */
  if (pdoc) fprintf(ftbl,"\nPFIR Coefficients:");
  if (outstyle==GC100) fprintf(fout,"\n# PFIR Coefficients:");
  for(j=0; j<32; j++) {
     if(pdoc&&((j%8)==0)) fprintf(ftbl,"\n%2d",2+(j/8)+i*8);
     c = chan[i].coef63[j];
     writec(2+(j/8)+i*8,16+2*(j%8), c & 0xff);
     writec(2+(j/8)+i*8,17+2*(j%8), (c>>8)&0xff);
     if(pdoc) fprintf(ftbl, " %2x %2x",c & 0xff, (c>>8) & 0xff);
    if(outstyle==GC100) {
	if (j%8==0) {
          fprintf(fout, "\n");
	  writeit_GC100(fout,2,2*(2+(j/8)+i*8));
	  fprintf(fout,"write_block 0xd010 \\\n");
	}
        fprintf(fout, " 0x%02x 0x%02x",c & 0xff, (c>>8) & 0xff);
    }
  }

  /* page 6 frequency control */
  if (pdoc) {
      fprintf(ftbl,"\nFrequency tuning page:\n");
      fprintf(ftbl,"%2d",6+i*8);
      fprintf(ftbl," %2x", chan[i].phase & 0xff);
      fprintf(ftbl," %2x", (chan[i].phase>>8) & 0xff);
      fprintf(ftbl," %2x", (chan[i].freq) & 0xff);
      fprintf(ftbl," %2x", (chan[i].freq>>8) & 0xff);
      fprintf(ftbl," %2x", (chan[i].freq>>16) & 0xff);
      fprintf(ftbl," %2x", (chan[i].freq>>24) & 0xff);
      fprintf(ftbl,"\n");
  }
  if (outstyle==GC100) {
      fprintf(fout,"\n# Frequency tuning page:\n");
      writeit_GC100(fout,2,2*(6+i*8));
      fprintf(fout,"write_block 0xd010 \\\n");
      fprintf(fout," 0x%02x", chan[i].phase & 0xff);
      fprintf(fout," 0x%02x", (chan[i].phase>>8) & 0xff);
      fprintf(fout," 0x%02x", (chan[i].freq) & 0xff);
      fprintf(fout," 0x%02x", (chan[i].freq>>8) & 0xff);
      fprintf(fout," 0x%02x", (chan[i].freq>>16) & 0xff);
      fprintf(fout," 0x%02x", (chan[i].freq>>24) & 0xff);
      fprintf(fout,"\n");
  }
  writec(6+i*8,16, chan[i].phase & 0xff);
  writec(6+i*8,17, (chan[i].phase>>8) & 0xff);
  writec(6+i*8,18, (chan[i].freq) & 0xff);
  writec(6+i*8,19, (chan[i].freq>>8) & 0xff);
  writec(6+i*8,20, (chan[i].freq>>16) & 0xff);
  writec(6+i*8,21, (chan[i].freq>>24) & 0xff);

  /* page 7 channel control */
  r[16]=(chan[i].cex&7) + ((chan[i].scex&1)<<3) + (((1-chan[i].enable)&1)<<7);
  r[17]=(chan[i].phasesync&7) + ((chan[i].freqsync&7)<<4);
  r[18]=(chan[i].ncosync&7) + ((chan[i].dithersync&7)<<4);
  r[19]=(chan[i].blankcnt&15) + ((chan[i].blanksync&7)<<4)+((chan[i].blanken&1)<<7);
  r[20]=(chan[i].deccountsync&7) + ((chan[i].flushsync&7)<<4);
  r[21]=(chan[i].deccount) & 0xff;
  r[22]=(((chan[i].deccount>>8) & 0xf)| (chan[i].gainsync<<4));
  r[23]=(chan[i].cicscale&0x3f) + ((chan[i].out20&1)<<6);
  r[24]=(chan[i].negctl&15) + ((chan[i].real2x&1)<<4) + ((chan[i].ionly<<5)) + ((chan[i].qonly<<6));
  r[25]=(chan[i].nosym21&1) + ((chan[i].qdlyqtr&1)<<1) + ((chan[i].idlyqtr&1)<<2)
	+ ((chan[i].coarse&7)<<4) + ((chan[i].test&1)<<7);
  r[26]=chan[i].nosym63
      +(chan[i].qdlyhlf<<1)+(chan[i].idlyhlf<<2)+(chan[i].ofselect<<3);
  r[27]=(chan[i].in&7) + ((chan[i].ifa&1)<<4) + ((chan[i].ifmt&3)<<4) +
		((chan[i].msbpol&1)<<7);
  r[28]=chan[i].peaksync +
		(chan[i].peakthresh<<3)+(chan[i].peakselect<<5);
  r[29]=0; /* address 29 is read only */
  r[30]=(chan[i].finegain)&0xff;
  r[31]=(chan[i].finegain>>8)&0xff;
  for(k=16;k<32;k++) 
      if(k!=29) writec(7+i*8,k,r[k]);
  if(pdoc) {
      fprintf(ftbl,"General controls:\n");
      fprintf(ftbl,"%2d",7+i*8);
      for(k=16;k<32;k++)
          if(k==29) fprintf(ftbl, "  -");
          else fprintf(ftbl, " %2x",r[k]);
      fprintf(ftbl,"\n");
   }
  if(outstyle==GC100) {
      fprintf(fout,"# General controls:\n");
      writeit_GC100(fout,2,2*(7+i*8));
      fprintf(fout,"write_block 0xd010 \\\n");
      for(k=16;k<32;k++) fprintf(fout, " 0x%02x",r[k]);
      fprintf(fout,"\n");
   }
   if (outstyle==INI) writeini(fout,7+i*8,r);
}

printres() { 
    int i, c, k;
    int r[32];

    /* filter coefficients */
    if(nosymres) c = nrestaps;
    else         c = nrestaps/2;
    for(i=0; i<256; i+=4) writerescoef(i);
    if (pdoc) {
	fprintf(ftbl, "\nResampler Coefficient pages:");
        for(i=0; i<c; i++) {
	  k = rescoef[i];
	  if((i%8)==0) fprintf(ftbl,"\n%2d",32+(i/8));
	  fprintf(ftbl," %2x %2x",k&0xff, (k>>8)&0x0f);
	}
	fprintf(ftbl,"\n");
    }
    if (outstyle==GC100) {
	fprintf(fout, "\n# ------ Resampler pages ------\n");
	fprintf(fout, "\n# Resampler Coefficient pages:");
        for(i=0; i<c; i++) {
	  k = rescoef[i];
	  if((i%8)==0) {
	      fprintf(fout,"\n");
	      writeit_GC100(fout,2,2*(32+(i/8)));
	      fprintf(fout,"write_block 0xd010 \\\n");
	  }
	  fprintf(fout," 0x%02x 0x%02x",k&0xff, (k>>8)&0x0f);
	}
	fprintf(fout,"\n");
    }

    /* resampler registers */
    writec(64,16, (r[16]=nochan + (sfiltcon<<2) + (ressync<<4)));
    writec(64,17, (r[17]=nrm + (nosymres<<6)));
    for(i=c=0; i<4; i++) c += (ochan[i].filterselect << (i+i));
    writec(64,18, (r[18]=c));
    writec(64,19, (r[19]=(gen.tag22<<6)+(round<<4)+finalshift));
    for(i=c=0; i<4; i++) c += (chan[i].outchan << (i+i));
    writec(64,20, (r[20]=c));
    for(i=c=0; i<4; i++) c += (chan[i].add_to << i);
    writec(64,21, (r[21]=c + (ratiosync<<4)));
    writec(64,22, (r[22]=resclkdiv));
    writec(64,23, (r[23]=ratiomap));
    if(pdoc) {
        fprintf(ftbl,"\nResampler control page:\n64");
        for(i=16; i<24;i++) fprintf(ftbl," %2x",r[i]);
        fprintf(ftbl,"\n");
    }
    if(outstyle==GC100) {
        fprintf(fout,"\n# Resampler control page:\n");
	writeit_GC100(fout,2,128);
	fprintf(fout,"write_block 0xd010 \\\n");
        for(i=16; i<24;i++) fprintf(fout," 0x%02x",r[i]);
        fprintf(fout,"\n");
    }
   if (outstyle==INI) writeini(fout,64,r);


    /* resample ratios */
    for(i=0;i<32;i++) r[i]=0;
    for(i=0; i<4; i++) {
      c=ochan[i].resampratio;
      for(k=0; k<4; k++) r[16+4*i+k]= (c>>(k*8)) & 0xff;
    }
    for(i=16; i<32; i++) writec(65,i,r[i]);
    if(pdoc) {
        fprintf(ftbl, "\nResampler ratio page:\n65");
        for(i=16; i<32; i++) fprintf(ftbl," %2x", r[i]);
        fprintf(ftbl,"\n");
    }
    if(outstyle==GC100) {
        fprintf(fout,"\n# Resampler ratio page:\n");
	writeit_GC100(fout,2,130);
	fprintf(fout,"write_block 0xd010 \\\n");
        for(i=16; i<32;i++) fprintf(fout," 0x%02x",r[i]);
        fprintf(fout,"\n");
    }
   if (outstyle==INI) writeini(fout,65,r);

}

/* various output formats for TI, GC100, and simulation */
writec(p,a,c)
    int p,a,c;
    {
    if(outstyle==TI) {
        writeit_TI(fout,p, a,c);
        return;
    }
    if(a<16) { /* a global register so it is easy */
       if(a==2) { /* direct write to the page register save the new page */
           page = (c>>1);
        }
    } else { /* check to see if we need to set the page first */
       if(page != p) writec(0,2,(p<<1));
    }
    if(outstyle==SIM  ) writeit_SIM(fout, a,c);
} 

/* version to write for 4 bit TI eval card */
writeit_TI(fp,p,a,c) 
    int a,c,p;
    FILE *fp;
    {
    int aa;
    if (a<8) {
      if(a==2) { /* writing to page register */
        page = c;
      } 
      fprintf(fp,"%x %x\n",a,c);
    } else {
        if (p != (page>>1)) writeit_TI(fp,0,2,p<<1);
	/* a paged register so check for a3 */
	if(((a>>3)&1)!=(page&1)) writeit_TI(fp,0,2,page^1);
	/* mapping that takes gc4017 a4 and makes it TIC6202 a3 */
	aa = (a&7) | ((a>>1)&8);
    	fprintf(fp, "%x %x\n", aa, c);
    }
}

/* version to write for gc100 tester */
writeit_GC100(fp, a,c) 
    int a,c;
    FILE *fp;
    {
    fprintf(fp, "write 0x%x 0x%02x\n", 0xd000+a, c);
    }

/* version to write for simulator */
writeit_SIM(fp, a,c) 
    int a,c;
    FILE *fp;
    {
    fprintf(fp, "writec 0x%x 0x%x\n", a, c);
    }

writeini(fp, page, regs) 
/* routine to write in a format convenient to echotek */
/* wish I had been faster a publishing a format so others would use it 
 * rather than creating so many variants 
 * alas, next time.
 */
int page, regs[]; 
FILE *fp; 

{
int i,j,k,aa, channel;
if(page==-1) {
fprintf(fp, "#----------------------------------------------------------------------------#\n");
fprintf(fp, "#-:                                                                        :-#\n");
fprintf(fp, "#-: Name          :                                                        :-#\n");
fprintf(fp, "#-: Created by    : Lee/Greg                                               :-#\n");
fprintf(fp, "#-: Creation Date :                                                        :-#\n");
fprintf(fp, "#-: Note          : All input in HEX unless specified otherwise            :-#\n");
fprintf(fp, "#-: Description   :                                                        :-#\n");
fprintf(fp, "#-:               :                                                        :-#\n");
fprintf(fp, "#-:               :                                                        :-#\n");
fprintf(fp, "#-:               :                                                        :-#\n");
fprintf(fp, "#-:               :                                                        :-#\n");
fprintf(fp, "#----------------------------------------------------------------------------#\n");
fprintf(fp, "\n");
fprintf(fp, "\n");
fprintf(fp, "#---------------------[ Board Wide Configuration ]---------------------------#\n");
fprintf(fp, "\n");
fprintf(fp, "#-: GC 214 Configuration Registers :-#\n");
fprintf(fp, "#-:--------------------------------:-#\n");
fprintf(fp, "\n");
fprintf(fp, "#-: Register Name        Register Value     Register Width :-*/\n");
fprintf(fp, "#-: -------------        --------------     -------------- :-*/\n");
fprintf(fp, "    GC214_CONFIG=            000000F0	    #-: 32 Bits    :-*/\n");
fprintf(fp, "    GC214_CLK_DLY=           00000255	    #-: 32 Bits    :-*/\n");
fprintf(fp, "\n");
fprintf(fp, "\n");
fprintf(fp, "\n");
fprintf(fp, "\n");
fprintf(fp, "#-: Gray Config. :-#\n");
fprintf(fp, "#-:--------------:-#\n");
fprintf(fp, "\n");
fprintf(fp, "#-: Register Name        Register Value     Register Width :-*/\n");
fprintf(fp, "#-: -------------        --------------     -------------- :-*/\n");
fprintf(fp, "    GLOBAL_RESET=            %02X		    #-:  8 Bits Always use Reset :-*/\n", regs[0]);
fprintf(fp, "    GENERAL_SYNC=            %02X		    #-:  8 Bits    :-*/\n", regs[4]);
fprintf(fp, "    COUNT_SYNC=              %02X		    #-:  8 Bits    :-*/\n", regs[5]);
fprintf(fp, "    COUNTER=                 %04X           #-: 16 Bits    :-*/\n", regs[6]+(regs[7]<<8));
}
if(page==64) { /* resampler */
fprintf(fp, "    N_CHANNELS_OUT=          %02X		    #-:  8 Bits    :-*/\n", regs[16]);
fprintf(fp, "    N_MULTIPLIES=            %02X		    #-:  8 Bits    :-*/\n", regs[17]);
fprintf(fp, "    FILTER_SELECT=           %02X		    #-:  8 Bits    :-*/\n", regs[18]);
fprintf(fp, "    FINAL_SHIFT=             %02X		    #-:  8 Bits    :-*/\n", regs[19]);
fprintf(fp, "    CHANNEL_MAP=             %02X		    #-:  8 Bits    :-*/\n", regs[20]);
fprintf(fp, "    ADD_TO=                  %02X		    #-:  8 Bits    :-*/\n", regs[21]);
fprintf(fp, "    RESAMPLER_CLOCK_DIVIDE=  %02X		    #-:  8 Bits    :-*/\n", regs[22]);
for(k=0; k<4; k++) {
for(aa=j=0; j<4; j++) aa += (regs[16+j+4*k]<<(j*8));
fprintf(fp, "    RATIO_%d=                 %08X       #-: 32 Bits    :-*/\n", k,aa);
}
}
if(page==98) { /* output */
fprintf(fp, "    TRISTATE_CTRL=           %02X		    #-:  8 Bits    :-*/\n", regs[16]);
fprintf(fp, "    OUTPUT_FORMAT=           %02X		    #-:  8 Bits    :-*/\n", regs[17]);
fprintf(fp, "    OUTPUT_MODE=             %02X		    #-:  8 Bits    :-*/\n", regs[18]);
fprintf(fp, "    OUTPUT_FRAME_CTRL=       %02X		    #-:  8 Bits    :-*/\n", regs[19]);
fprintf(fp, "    OUTPUT_WORD_SIZE=        %02X		    #-:  8 Bits    :-*/\n", regs[20]);
fprintf(fp, "    OUTPUT_CLOCK_CTRL=       %02X		    #-:  8 Bits    :-*/\n", regs[21]);
fprintf(fp, "    SERIAL_MUX_CTRL=         %02X		    #-:  8 Bits    :-*/\n", regs[22]);
fprintf(fp, "    OUTPUT_TAG_A=            %02X		    #-:  8 Bits    :-*/\n", regs[23]);
fprintf(fp, "    OUTPUT_TAG_B=            %02X		    #-:  8 Bits    :-*/\n", regs[24]);
fprintf(fp, "    OUTPUT_TAG_C=            %02X		    #-:  8 Bits    :-*/\n", regs[25]);
fprintf(fp, "    OUTPUT_TAG_D=            %02X		    #-:  8 Bits    :-*/\n", regs[26]);
fprintf(fp, "\n");
fprintf(fp, "\n");
fprintf(fp, "\n");
fprintf(fp, "\n");
}
if((page>-1) && (page < 32))
if((page&0x7) == 7)  { /* channel control */
fprintf(fp, "#---------------------[ Channel Wide Configuration ]-------------------------#\n");
fprintf(fp, "\n");
fprintf(fp, "#-: Channel Specifier              :-#\n");
fprintf(fp, "#-: Valid Channel Specifier = 1- 4 :-#\n");
fprintf(fp, "\n");
channel = (page>>3);
fprintf(fp, "GC214_CHANNEL= %d\n",(page>>3)+1);		
fprintf(fp, "\n");
fprintf(fp, "#-: Register Name        Register Value     Register Width :-*/\n");
fprintf(fp, "#-: -------------        --------------     -------------- :-*/\n");
fprintf(fp, "    CH_RESET=                %02X		    #-:  8 Bits    :-*/\n", regs[16]);
fprintf(fp, "    FREQ_SYNC=               %02X		    #-:  8 Bits    :-*/\n", regs[17]);
fprintf(fp, "    NCO_SYNC=                %02X		    #-:  8 Bits    :-*/\n", regs[18]);
fprintf(fp, "    ZPAD_MODE_CTRL=          %02X		    #-:  8 Bits    :-*/\n", regs[19]);
fprintf(fp, "    DEC_FLUSH_SYNC=          %02X		    #-:  8 Bits    :-*/\n", regs[20]);
fprintf(fp, "    DEC_RATIO=               %04X             #-: 16 Bits    :-*/\n", regs[21]+(regs[22]<<8));
fprintf(fp, "    CIC_SCALE=               %02X		    #-:  8 Bits    :-*/\n", regs[23]);
fprintf(fp, "    SPLIT_IQ=                %02X		    #-:  8 Bits    :-*/\n", regs[24]);
fprintf(fp, "    CFIR=                    %02X		    #-:  8 Bits    :-*/\n", regs[25]);
fprintf(fp, "    PFIR=                    %02X		    #-:  8 Bits    :-*/\n", regs[26]);
fprintf(fp, "    INPUT=                   %02X		    #-:  8 Bits    :-*/\n", regs[27]);
fprintf(fp, "    PEAK_CTRL=               %02X		    #-:  8 Bits    :-*/\n", regs[28]);
fprintf(fp, "    PEAK_COUNT=              %02X		    #-:  8 Bits    :-*/\n", regs[29]);
fprintf(fp, "    FINE_GAIN=               %04X           #-: 16 Bits    :-*/\n", chan[channel].finegain);
fprintf(fp, "    PHASE=                   %04X           #-: 16 Bits    :-*/\n", chan[channel].phase);
fprintf(fp, "    FREQUENCY=         	     %08X       #-: 32 Bits    :-*/\n", chan[channel].freq);
fprintf(fp, "\n");
fprintf(fp, "#-: Coefficients Section :-#\n");
fprintf(fp, "TOTAL_CFIR_COEFF= 11\n");
for(j=0; j<11;j++) fprintf(fp, "  %.0lf\n", chan[channel].coef21[j]);
fprintf(fp, "CFIR_END\n");
fprintf(fp, "\n");
fprintf(fp, "TOTAL_PFIR_COEFF= 32\n");
for(j=0; j<32;j++) fprintf(fp, "  %.0lf\n", chan[channel].coef63[j]);
fprintf(fp, "PFIR_END\n");
fprintf(fp, "\n");
fprintf(fp, "GC214_CHANNEL_END\n", regs[00]);
fprintf(fp, "\n");
fprintf(fp, "\n");
}
if(page==65) {
fprintf(fp, "\n");
fprintf(fp, "\n");
fprintf(fp, "#-----------------------[ Resampler Coefficients  ]--------------------------#\n");
fprintf(fp, "TOTAL_RESAMPLER_COEFF= 256\n");
for(i=0; i<256;i++) fprintf(fp, "  %.0lf\n", rescoef[i]);
fprintf(fp, "RESAMPLER_END\n");
for(i=0; i<23; i++) fprintf(fp, "\n");
}
}
/* breakout multi channel modes by having the user explictly state
 * which he is using to simplify checking
 * add check for counter for multiple of decimation ratio (?)
 * is wideword uP done by enabling parallel ?
 */

#define RESOVER 0
#define RESBIGLOAD 0.9
#define RESSHIFT 26
#define NRM (nrm+1) 
#define uP 0
#define SERIAL 1
#define NIBBLE 2
#define PARALLEL 3

double exp2(i)
int i;
{
    int k;
    double x;
    x=1.0;
    if (i<0) for(k=0;k<-i;k++) x *= 0.5;
    else   for(k=0;k<i;k++) x *= 2.0;
    return(x);
}

/*   check for reasonableness of the input parameters */
sanity(report, gen) 
   int report;
   GEN *gen;
   {
    int i,j, err, maxrmem, nphases, src[4];
    int off, oc, oca[4], addto, myfmt;
    double sumave,summin,summax;
    double gain, sum, temp, resfl;
    double rescompute;
    CHAN *cp;
    double gaino[4];

    if(report<0) return;
    /* check for input blanking compatibility */
    for(i=0; i<4; i++) {
	if(chan[i].enable==0) continue;
        for(j=i+1; j<4; j++) {
	   if(chan[j].enable==0) continue;
	   if(chan[i].in == chan[j].in) { /* they use same input */
	   /* better be the same blanking */
	   /* future can relax input format restriction to this as well */
	       if((chan[i].blanken != chan[j].blanken) ||
	          (chan[i].blankcnt != chan[j].blankcnt) )
			{
		  fprintf(stderr, "Zero padding must match when two %s",
		      "channel use the same input port \n");
		  fprintf(ftbl, "Zero padding must match when two %s",
		      "channel use the same input port \n");
		      }
	    }
	}
    }

    for(i=0; i<4; i++) {
	if(chan[i].enable!=0) { myfmt = chan[i].ifmt; break;}
    }
    chan[0].ifmt = myfmt;
    for(i=0; i<4; i++) { 
	if((chan[i].ifmt != myfmt) && (chan[i].enable!=0))
	{
	fprintf(stderr, "Simulator requires all active channels to use the same format\n");
	fprintf(ftbl, "Simulator requires all active channels to use the same format\n");
	}
    }

    for(i=0; i<4; i++) {src[i]=-1; gaino[i]=-1; oca[i]=-1;}
    err = rescompute = 0;
    for(i=0; i<4; i++) {
       if(chan[i].enable!=0) {
	   if(report) fprintf(stderr,"\nChecks for input channel %d\n",i);
	   if(report) fprintf(ftbl,"\nChecks for input channel %d\n",i);
	   cp = &chan[i];
	   /* is output channel valid */
	   oca[cp->outchan] = 1;
	   if(cp->outchan > nochan) {
	       fprintf(stderr, "==> Input channel %d is directed to output channel %d\n",
		  i, cp->outchan);
		fprintf(stderr, "==> But you only declared output channels up to %d valid\n",
		  nochan);
	       fprintf(ftbl, "==> Input channel %d is directed to output channel %d\n",
		  i, cp->outchan);
		fprintf(ftbl, "==> But you only declared output channels up to %d valid\n",
		  nochan);
		err=1;
	    }
	   /* constant exponent and 20 bit mixer output */
	   if(cp->out20 && cp->cex<4 && cp->scex) {
		fprintf(stderr, "==> Shift must be greater %s",
		"than 3 when using 20 bit output from mixer\n");
		fprintf(ftbl, "==> Shift must be greater %s",
		"than 3 when using 20 bit output from mixer\n");
		err = 1;
	    }

	   /* normal gain through CIC */
	   temp = cp->deccount+1;
	   if(cp->real2x) temp = temp/2.0;
	   gain = temp*temp*temp*temp*temp;
	   if(cp->scex) gain *= exp2(cp->cex);
	   gain *= exp2((cp->cicscale&0x7) + ((cp->cicscale>>3)&0x7)*6-62);
	   if(report) fprintf(stderr, "CIC Gain is %lf\n", gain);
	   if(report) fprintf(ftbl, "CIC Gain is %lf\n", gain);
	   if(gain>1.0) {
	       fprintf(stderr, "==> Excessive gain through CIC.  Gain is %.1lf\n", gain);
	       fprintf(stderr, "==> Full scale inputs will cause nonsaturated overflow\n");
	       fprintf(ftbl, "==> Excessive gain through CIC.  Gain is %.1lf\n", gain);
	       fprintf(ftbl, "==> Full scale inputs will cause nonsaturated overflow\n");
	       err = 0;
	    }
	    
	    /* gain at input to FIR1 */
	    gain *= exp2(cp->coarse);
	    if(report) fprintf(stderr, "Gain at input CFIR is %lf\n", gain);
	    if(report) fprintf(ftbl, "Gain at input CFIR is %lf\n", gain);
	    if(gain>1.0) {
		fprintf(stderr, "Signal boosted at CFIR input. Gain is %.1lf\n", gain);
		fprintf(ftbl, "Signal boosted at CFIR input. Gain is %.1lf\n", gain);
		err = 0;
	    }
	    
	    /* gain at input to FIR2 */
	    sum = cp->coef21[10];
	    for(j=0; j<10; j++) sum += (cp->coef21[j]*(2-cp->nosym21));
	    gain *= sum/65536.0;
	    if(report) fprintf(stderr, "CFIRout Gain is %lf\n", gain);
	    if(report) fprintf(ftbl, "CFIRout Gain is %lf\n", gain);
	    if(gain>1.0) {
		fprintf(stderr, "Signal boosted at CFIR output. Gain is %.1lf\n", gain);
		fprintf(ftbl, "Signal boosted at CFIR output. Gain is %.1lf\n", gain);
		err=0;
	    }

	    /* gain at input to resampler */
	    sum = cp->coef63[31];
	    for(j=0; j<31; j++) sum += (cp->coef63[j]*(2-cp->nosym63));
	    gain *= sum/65536.0*(cp->finegain/1024.0);
	    if(report) fprintf(stderr, "PFIRout Gain is %lf\n", gain);
	    if(report) fprintf(ftbl, "PFIRout Gain is %lf\n", gain);
	    if(gain>1.0) {
		fprintf(stderr, "Signal boosted at PFIR output. Gain is %.1lf\n", gain);
		fprintf(ftbl, "Signal boosted at PFIR output. Gain is %.1lf\n", gain);
		err=0;
	    }
	    gaino[cp->outchan] = gain;

	    /* data rate from channel to resampler */
	    if((4*(cp->deccount+1))<=(8-nochan)*(resclkdiv+1)*(nochan+1)) {
	       fprintf(stderr, "==> Resampler clock division too high for this case\n");
	       fprintf(stderr, "==> Maximum recommended value is %4.0lf\n",
	           (((double) (4*(cp->deccount+1)))/(nochan+1)/(8.0-nochan) - 0.98));
	       fprintf(ftbl, "==> Resampler clock division too high for this case\n");
	       fprintf(ftbl, "==> Maximum recommended value is %4.0lf\n",
	           (((double) (4*(cp->deccount+1)))/(nochan+1)/(8.0-nochan) - 0.98));
	    }
	    /* resampler input data memory limitations */
	    if     (nochan==0) maxrmem = 63;
	    else if(nochan==1) maxrmem = 31;
	    else               maxrmem = 15;
	    if(NRM > maxrmem) { /* out of input data memory */
	       fprintf(stderr, "==> Resampler filter too long at %d multiplies\n", NRM);
	       fprintf(stderr, "==> Must be less than 64,32, or 16 respectively %s",
		  "for 1,2, or more channels\n");
	       fprintf(ftbl, "==> Resampler filter too long at %d multiplies\n", NRM);
	       fprintf(ftbl, "==> Must be less than 64,32, or 16 respectively %s",
		  "for 1,2, or more channels\n");
		err = 1;
	    }
	    /* assume no split for now */
	    resfl = 256.0 * (2.0-nosymres);
	    /* resampler filters require exact filter length */
	    nphases = ndelay;
	    if(nphases>64) nphases=64;
	    if(nrestaps != nphases*NRM) {
	       fprintf(stderr, "==> A resampler filter MUST be an exact length\n");
	       fprintf(stderr, "==> For nrm=%d this means %d\n",nrm, nphases*NRM);
	       fprintf(stderr, "==> You have %d taps\n",nrestaps);
	       fprintf(ftbl, "==> A resampler filter MUST be an exact length\n");
	       fprintf(ftbl, "==> For nrm=%d this means %d\n",nrm, nphases*NRM);
	       fprintf(ftbl, "==> You have %d taps\n",nrestaps);
	       err = 1;
	    }
	    /* resampler computation rates must be within bounds */
	    /* but do this check at the end */
	    /* for now add up the compute cycles */
	    if(cp->add_to==0)
	    rescompute += ((((double) (NRM + RESOVER))*(1<<RESSHIFT))/ochan[cp->outchan].resampratio)/
	    	(8.0*(cp->deccount+1));

     }}

     for(i=0; i<=nochan; i++) if(oca[i]!=1) {
	 fprintf(stderr,
     "Warning: %d output channels declared but output %d has no data input\n", nochan, i);
	 fprintf(ftbl,
     "Warning: %d output channels declared but output %d has no data input\n", nochan, i);
     }
     fprintf(stderr,"\n\nResampler configuration:\nQtaps = %d, Nmult = %d, Ndelay = %d, Nfilters = %d\n",
	 nrestaps, NRM, nphases, sfiltcon+1);
     fprintf(ftbl,"\n\nResampler configuration:\nQtaps = %d, Nmult = %d, Ndelay = %d, Nfilters = %d\n",
	 nrestaps, NRM, nphases, sfiltcon+1);

     /* Too much compute load for the resampler ? */
     temp = rescompute * (resclkdiv+1);
     if(report) fprintf(stderr, "\nResampler Compute load is: %lf\n", temp);
     if(report) fprintf(ftbl, "\nResampler Compute load is: %lf\n", temp);
     if(temp > 1.0) {
	fprintf(stderr, "==> Resampler computation is overloaded.\n");
	fprintf(stderr, "==> Load factor is %lf\n", temp);
	fprintf(ftbl, "==> Resampler computation is overloaded.\n");
	fprintf(ftbl, "==> Load factor is %lf\n", temp);
	if(rescompute>1.0) {
     	    temp = ((int) rescompute) - 1;
	    if(temp>0) 
	       fprintf(stderr, "==> Maximum value for res_clk_div is %d\n",temp);
	    else fprintf(stderr,"==> You'll need to reduce nm\n");
	    if(temp>0) 
	       fprintf(ftbl, "==> Maximum value for res_clk_div is %d\n",temp);
	    else fprintf(ftbl,"==> You'll need to reduce nm\n");
	}
     } else if( (temp > RESBIGLOAD) && ((maxrmem-NRM) < 2)) {
	fprintf(stderr, "==> Warning: Near maximum compute load and maximum memory depth\n");
	fprintf(stderr, "==> May cause resampler input fifo overflows\n");
	fprintf(ftbl, "==> Warning: Near maximum compute load and maximum memory depth\n");
	fprintf(ftbl, "==> May cause resampler input fifo overflows\n");
    }

    /* Check for parallel inputs to resampler */
    for(i=0; i<4; i++) {
	if(chan[i].enable==0) continue;
	addto = 0;
	if(chan[i].real2x)  { /* real2x implies ionly or qonly must be set */
	    if(chan[i].add_to) { /* real2x and add_to means complex in 4inchan->1outchan */
		addto = 1;
		err |= checkreal2x(0,2,report);
		err |= checkreal2x(1,3,report);
		break;
	    }
	    if(addto==0) {
	        if(chan[i].ionly==1) err |= checkreal2x(i,i+1+chan[i].add_to,report);
	        else 	 err |= checkreal2x(i-1,i,report);
	    }
	}
	if( src[chan[i].outchan] < 0) src[chan[i].outchan] = i;
	else checkpair(src[chan[i].outchan], i);
    }

    /* checks on output channels */
    if(sfiltcon!=0) 
    {
	fprintf(stderr, "Warning. Checker does not check %s", 
	"multiple resampler filter coefficients\n");
	fprintf(ftbl, "Warning. Checker does not check %s", 
	"multiple resampler filter coefficients\n");
    } else {


    /* resampler filter gains */
    for(off=0; off<nphases; off++) {
	for(i=sum=0; i<NRM; i++) sum += rescoef[i*nphases+off];
	if(off==0) sumave=summin=summax=sum;
	else {
	    sumave += sum;
	    if(sum<summin) summin = sum;
	    if(sum>summax) summax = sum;
	}
    }
    sumave /= nphases;
    if(report) fprintf(stderr,"Resampler filter gain variations min %lf ave %lf max %lf\n",
	summin, sumave, summax);
    if(report) fprintf(ftbl,"Resampler filter gain variations min %lf ave %lf max %lf\n",
	summin, sumave, summax);

    /* gain at resampler output */
    for(oc = 0; oc <= nochan; oc++) {
	gain = gaino[oc];
	gain *= summax / ((double) (1<<(15-finalshift)));
	if(report) fprintf(stderr, "Overall gain for output %d is %lf\n", oc, gain);
	if(report) fprintf(ftbl, "Overall gain for output %d is %lf\n", oc, gain);
	if(gain>1.0) {
	    fprintf(stderr, "Signal for outchannel %d boosted at output. Gain is %.1lf\n", 
	        oc, gain);
	    fprintf(ftbl, "Signal for outchannel %d boosted at output. Gain is %.1lf\n", 
	        oc, gain);
	    err=1;
    	}
	/* filter selection */
	if(ochan[oc].filterselect > sfiltcon) {
	    fprintf(stderr, "Filter selection (filter_sel_%d) %s", oc,
	    "exceeds number of filters specified (nf)\n");
	    fprintf(ftbl, "Filter selection (filter_sel_%d) %s", oc,
	    "exceeds number of filters specified (nf)\n");
	    err = 1;
	}
    }
    }
    /* now check general registers and outputs */
    checkgen(gen);
}

checkreal2x(i,j,msg)
    int i,j,msg;
    {
    int err;
    err = 0;
    if(i<0) {
	fprintf(stderr,"==> Ionly needs to precede Qonly channel\n");
	fprintf(ftbl,"==> Ionly needs to precede Qonly channel\n");return(1);
    }
    if(j>3) {
	fprintf(stderr,"==> Qonly needs to follow Ionly channel\n");
	fprintf(ftbl,"==> Qonly needs to follow Ionly channel\n");return(1);
    }
    if(chan[i].enable==0) {
	fprintf(stderr,"==> Channel %d needs to be enabled\n", i); 
	fprintf(ftbl,"==> Channel %d needs to be enabled\n", i); err = 1; 
    }
    if(chan[j].enable==0) {
	fprintf(stderr,"==> Channel %d needs to be enabled\n", j); 
	fprintf(ftbl,"==> Channel %d needs to be enabled\n", j); err = 1; 
    }
    if(chan[i].ionly!=1) { 
      fprintf(stderr,"==> Ionly for channel %d needs to be on\n", i); 
      fprintf(ftbl,"==> Ionly for channel %d needs to be on\n", i); err = 1; 
    }
    if(chan[j].qonly!=1) { 
      fprintf(stderr,"==> Qonly for channel %d needs to be on\n", j); 
      fprintf(ftbl,"==> Qonly for channel %d needs to be on\n", j); err = 1; 
    }
    err |= checkpair(i,j,msg);
    return(err);
    }

checkpair(i,j,msg)
    int i,j,msg;
    {
    CHAN *a,*b;
    int err, k;

    if(i<0) {
      fprintf(stderr,"==> Ionly needs to precede Qonly channel\n");
      fprintf(ftbl,"==> Ionly needs to precede Qonly channel\n");return(1);
    }
    if(j>3) {
      fprintf(stderr,"==> Qonly needs to follow Ionly channel\n");
      fprintf(ftbl,"==> Qonly needs to follow Ionly channel\n");return(1);
    }
    if(chan[i].enable==0) {
      fprintf(stderr,"==> Need to enable channel %d\n",i); 
      fprintf(ftbl,"==> Need to enable channel %d\n",i); return(1);
    }
    if(chan[j].enable==0) {
      fprintf(stderr,"==> Need to enable channel %d\n",j); 
      fprintf(ftbl,"==> Need to enable channel %d\n",j); return(1);
    }
    /* other checks for channels paired  */
    a = &chan[i];
    b = &chan[j];
    err = 0;
    err |= ckmatch(a->blanken,b->blanken,"zpad_en");
    if(a->blanken) err |= ckmatch(a->blankcnt,b->blankcnt,"nzeros");
    if(a->blanken) err |= ckmatch(a->blanksync,b->blanksync,"zpad_sync");
    err |= ckmatch(a->scex,b->scex,"use_shift");
    if(a->scex) err |= ckmatch(a->cex,b->cex,"shift");
    err |= ckmatch(a->cicscale,b->cicscale,"big_scale and scale");
    err |= ckmatch(a->coarse,b->coarse,"coarse");
    err |= ckmatch(a->deccount,b->deccount,"dec");
    err |= ckmatch(a->deccountsync,b->deccountsync,"dec_sync");
    err |= ckmatch(a->dither,b->dither,"dither");
    if(a->dither) err |= ckmatch(a->dithersync,b->dithersync,"dither_sync");
    err |= ckmatch(a->flushsync,b->flushsync,"flush_sync");
    err |= ckmatch(a->freq,b->freq,"freq");
    err |= ckmatch(a->ifa,b->ifa,"ab_sel");
    err |= ckmatch(a->ifmt,b->ifmt,"in_fmt");
    if(a->add_to == 0) 	/* if input is complex then we use separate ports for I and Q */
    err |= ckmatch(a->in,b->in,"input_sel");
    err |= ckmatch(a->msbpol,b->msbpol,"msb_pol");
    err |= ckmatch(a->ncosync,b->ncosync,"nco_sync");
    err |= ckmatch(a->nosym21,b->nosym21,"no_sym_cfir");
    err |= ckmatch(a->nosym63,b->nosym63,"no_sym_pfir");
    err |= ckmatch(a->out20,b->out20,"mix20b");
    err |= ckmatch(a->outchan,b->outchan,"chan_map_X");
    err |= ckmatch(a->real2x,b->real2x,"splitiq");
    for(k=0; k<21; k++) {
	if(a->coef21[k]!=b->coef21[k]) {
	    err=1;
	    fprintf(stderr,"%s cfir coefficients must match tap %d\n",
	     "Paired channels", k);
	    fprintf(ftbl,"%s cfir coefficients must match tap %d\n",
	     "Paired channels", k);
	}
    }
    for(k=0; k<63; k++) {
	if(a->coef63[k]!=b->coef63[k]) {
	    err=1;
	    fprintf(stderr,"%s pfir coefficients must match tap %d\n",
	     "Paired channels", k);
	    fprintf(ftbl,"%s pfir coefficients must match tap %d\n",
	     "Paired channels", k);
	}
    }
    return(err);
}

ckmatch(a,b,str)
  int a,b; char *str;
  {
   if(a!=b) {
      fprintf(stderr, "==> %s must match when channels are paired\n",str);
      fprintf(ftbl, "==> %s must match when channels are paired\n",str);
      return(1);
    }
    return(0);
}



checkgen(gen) 
   GEN *gen;
   {
   int ncout;
   int i, async;
   double osamplerate;
   int stream[4];
   double csrate[4], trate;
   int nout,packed, cmax, minwpf;
   double csamplerate, framerate, temp, maxcsrate;

   /* general registers */
   if(gen->cklossdetect != 0) {
     fprintf(stderr, "Warning: ck_loss_detect should normally be set to zero\n");
     fprintf(stderr, "Warning: ck_loss_detect should normally be set to zero\n");
     }
   if(gen->ck2xtest != 0) {
      fprintf(stderr, 
	"Warning: ck_2x_test should normally be set to zero\n");
      fprintf(ftbl, 
	"Warning: ck_2x_test should normally be set to zero\n");
     }
   if(gen->ck2xen != 0) {
     fprintf(stderr, 
	"Warning: ck_2x_en should normally be set to zero\n");
     fprintf(ftbl, 
	"Warning: ck_2x_en should normally be set to zero\n");
   }
   if(gen->lvds !=0 ) { /* all inputs are differential.  
	Be sure all active channels are programmed to expect differential. */
	for(i=0; i<4; i++) {
	    if(chan[i].enable==0) continue;
	    if((chan[i].in != 4) && (chan[i].in != 5) ) {
	      fprintf(stderr, 
	"Error: lvds must be programmed both globally (address 0) and\n");
	      fprintf(stderr, 
	    "for each active channel (address 27 pages 7,15,23, and 31)\n");
	      fprintf(ftbl, 
	"Error: lvds must be programmed both globally (address 0) and\n");
	      fprintf(ftbl, 
	    "for each active channel (address 27 pages 7,15,23, and 31)\n");
	    }
	}
    }

   if(gen->counttest != 0) {
       fprintf(stderr, 
	"Warning: count_test should normally be set to zero\n");
       fprintf(ftbl, 
	"Warning: count_test should normally be set to zero\n");
	}

    /* complex output sample rate is */
    for(i=0; i<4; i++) csrate[i] = 0;
    ncout=0;
    for(i=0; i<4; i++) {
       if((chan[i].enable!=0) && (chan[i].add_to==0) && (chan[i].qonly==0)) {
	    ncout++; /* count number of output channels with something in them. */
	    csrate[chan[i].outchan] += 
		((double) ((1<<RESSHIFT))/
		(double) ochan[chan[i].outchan].resampratio)
	       * (1.0+chan[i].real2x) / (4.0*(chan[i].deccount+1));
	}
    }
    fprintf(stderr, "outrate %lf %lf %lf %lf\n", csrate[0], csrate[1],
	    csrate[2], csrate[3]);
    fprintf(ftbl, "outrate %lf %lf %lf %lf\n", csrate[0], csrate[1],
	    csrate[2], csrate[3]);
	for(i=osamplerate = 0; i<4; i++) osamplerate += csrate[i];
    osamplerate *= (2.0-realout);
    framerate = 1.0/(1.0+gen->ckdiv)/(4+4*gen->bnperword)/(1+gen->fl);
    /* if a slave we dont really know the frame rate but we can bound it
     * the real frame length must be at least (fl+1) + (wordsperframe+1)
     */
    if(gen->master==0) framerate = 1.0/(1.0+gen->ckdiv)/
	(4+4*gen->bnperword)/ (1+gen->fl+gen->wordsperframe+1);
    if(gen->os!=PARALLEL) {
       fprintf(stderr, "Frame rate %lf\n", framerate);
       fprintf(ftbl, "Frame rate %lf\n", framerate);
       }


    for(nout=i=0; i<4; i++) /* number of output channels */
       if((chan[i].enable!=0) && (chan[i].outchan>=nout)) nout = chan[i].outchan;
    if((nout>ncout-1) ) {
      fprintf(stderr, 
    "Warning the maximum output channel number assigned > number of output channels\n");
      fprintf(ftbl, 
    "Warning the maximum output channel number assigned > number of output channels\n");
    }
    if(gen->ni!=gen->no) {
       fprintf(stderr, 
    	"Warning unusual output sequence specified.  %s", 
	"Normally output_order = nserial\n");
       fprintf(ftbl, 
    	"Warning unusual output sequence specified.  %s", 
	"Normally output_order = nserial\n");
	}

    /* are the channels synchronous ? */
    for(async=trate= i=0; i<4; i++) {
	if(csrate[i]) {
	if(trate==0) trate = csrate[i];
	    else if(trate!=csrate[i]) async=1;
        }}
    if(async && (gen->ni!=0)) {
      fprintf(stderr, 
          "Error: With asynchronous output channels nserial must be 0\n");
      fprintf(ftbl, 
          "Error: With asynchronous output channels nserial must be 0\n");
    }
    if(async && (gen->tagen==0)) {
      fprintf(stderr,
	  "Error: With asynchronous output channels tags need to be on to understand the output\n");
      fprintf(ftbl,
	  "Error: With asynchronous output channels tags need to be on to understand the output\n");
    }

    for(maxcsrate=i=0; i<4; i++) if(maxcsrate<csrate[i]) maxcsrate=csrate[i];

    if(gen->os!=PARALLEL) {
	minwpf = (gen->blocklength+1)*(2-realout);
	if(gen->os==SERIAL) minwpf /= (1+gen->no);
	if((minwpf != (gen->wordsperframe+1))) {
	fprintf(stderr, "Warning: words_per_frame is normally %d for block_size %d\n",
	   minwpf-1, gen->blocklength);
	fprintf(ftbl, "Warning: words_per_frame is normally %d for block_size %d\n",
	   minwpf-1, gen->blocklength);
       }
    }
	    

    /* uP output checking */
    if(gen->os == uP) {
   	if(gen->enrdy == 0) {
	    fprintf(stderr, 
		"Warning: Need rdy enabled if it is used as an interrupt\n");
	    fprintf(ftbl, 
		"Warning: Need rdy enabled if it is used as an interrupt\n");
	}
        if(gen->link != 0) {
		fprintf(stderr, "Error: link must be zero in uP mode\n");
		fprintf(ftbl, "Error: link must be zero in uP mode\n");
	}
        if(gen->nibble != 0) {
		fprintf(stderr, "Error: nibble must be zero in uP mode\n");
		fprintf(ftbl, "Error: nibble must be zero in uP mode\n");
	}
        if(gen->blocklength != 3)  {
		fprintf(stderr, "Error: block_size must be 3 in uP mode\n");
		fprintf(ftbl, "Error: block_size must be 3 in uP mode\n");
	}

    /* serial output checking */
    } else if (gen->os == SERIAL) {
        if(framerate < maxcsrate) {
	  fprintf(stderr, "Error: frame rate must not %s ",
          "be less than the highest output channel rate\n");
	  fprintf(stderr, "Error: frame rate must not %s ",
          "be less than the highest output channel rate\n");
	}
        if((gen->master==0) && (gen->fl)) {
	  fprintf(stderr, 
		"Error: TDM slaves frame_length must be greater than 0\n");
	  fprintf(ftbl, 
		"Error: TDM slaves frame_length must be greater than 0\n");
	}
        if(gen->link+gen->nibble+gen->parallel) {
	  fprintf(stderr, 
        "Error: link,nibble, and parallel must be disabled in serial mode\n");
	  fprintf(ftbl, 
        "Error: link,nibble, and parallel must be disabled in serial mode\n");
        }
    if(gen->bnperword < (round+2+gen->tagen)) {
      fprintf(stderr,
	"Resampler round, output bits_per_word, and tagen aren't consistent\n");
      fprintf(ftbl,
	"Resampler round, output bits_per_word, and tagen aren't consistent\n");
      }
      /* now check for enough space on serial bus for words */
      /* for 2 or four streams we need to support the highest channel number assigned */
      /* probably can let wordsperframe be larger w/o harm but not sure */
      if(gen->no) cmax = (nout+1)*(2-realout);
      else   cmax = ncout*(2-realout);
      if((gen->wordsperframe+1) > cmax) {
	fprintf(stderr, 
	"Warning: words_per_frame larger than required.  Required is %d\n", cmax-1);
	fprintf(ftbl, 
	"Warning: words_per_frame larger than required.  Required is %d\n", cmax-1);
      }
      if( (gen->wordsperframe+1) < cmax ) {
	fprintf(stderr,
	    "Error: words_per_frame must be at least %d to get your data out\n",cmax-1);
	fprintf(ftbl,
	    "Error: words_per_frame must be at least %d to get your data out\n",cmax-1);
      }
      if(cmax/(gen->no+1) > (gen->fl+1)) {
	fprintf(stderr, 
	    "Error: Not enough slots on bus. Increase frame_length. Min %d\n", cmax/(gen->no+1));
	fprintf(ftbl, 
	    "Error: Not enough slots on bus. Increase frame_length. Min %d\n", cmax/(gen->no+1));
	}

    /* now check that the streams go someplace */
    for(i=0; i<4; i++) stream[i]=0;
    for(i=0; i<4; i++) stream[chan[i].outchan]=1;
    /* for each pin */
	if(gen->ensa) {
	    if(stream[gen->ssa]==0) {
	       fprintf(stderr, 
	      "Warning: Serial port p0 enabled but no active stream selected\n");
	       fprintf(ftbl, 
	      "Warning: Serial port p0 enabled but no active stream selected\n");
	    }
	    stream[gen->ssa] = 3;
	}
	if(gen->ensb) {
	    if(stream[gen->ssb]==0) {
	       fprintf(stderr, 
	      "Warning: Serial port p1 enabled but no active stream selected\n");
	       fprintf(ftbl, 
	      "Warning: Serial port p1 enabled but no active stream selected\n");
	    }
	    stream[gen->ssb] = 3;
	}
	if(gen->ensc) {
	    if(stream[gen->ssc]==0) {
	       fprintf(stderr, 
	      "Warning: Serial port p2 enabled but no active stream selected\n");
	       fprintf(ftbl, 
	      "Warning: Serial port p2 enabled but no active stream selected\n");
	    }
	    stream[gen->ssc] = 3;
	}
	if(gen->ensd) {
	    if(stream[gen->ssd]==0) {
	       fprintf(stderr, 
	      "Warning: Serial port p3 enabled but no active stream selected\n");
	       fprintf(ftbl, 
	      "Warning: Serial port p3 enabled but no active stream selected\n");
	    }
	    stream[gen->ssd] = 3;
	}
	for(i=0; i<4; i++) if(stream[i]==1) {
	   fprintf(stderr,
	      "Warning: Stream %d created but not output\n", i);
	   fprintf(ftbl,
	      "Warning: Stream %d created but not output\n", i);
	   }

    } else if (gen->os == NIBBLE) { /* LINK and nibble output checking */
        if(gen->parallel != 0)  {
		fprintf(stderr, "Error: parallel must be zero in parallel mode\n");
		fprintf(ftbl, "Error: parallel must be zero in parallel mode\n");
	}
        if((gen->master==0) && (gen->fl)) {
	   fprintf(stderr, 
		"Error: TDM slaves frame_length must be greater than 0\n");
	   fprintf(ftbl, 
		"Error: TDM slaves frame_length must be greater than 0\n");
	}
        if(gen->no != 0) {
	   fprintf(stderr, 
		"Error: nserial must be 0 in LINK or nibble modes\n");
	   fprintf(ftbl, 
		"Error: nserial must be 0 in LINK or nibble modes\n");
	}
	if(gen->bnperword==0) packed = 1;
	else if(gen->bnperword==1) packed = 0;
	else {
	    fprintf(stderr, 
	"Error: link and nibble modes require bits_per_word to be 0 or 1\n");
	    fprintf(ftbl, 
	"Error: link and nibble modes require bits_per_word to be 0 or 1\n");
	}

	if(gen->link &! (gen->bnperword ^ gen->wordsperframe)) {
	  fprintf(stderr,
            "Error: LINK must either be 2 16 bit words per frame %s",
	    "(bits_per_word=0,words_per_frame=1)\n              or be 1 %s",
            "32 bit word per frame (bits_per_word=1,words_per_frame=0)\n");
	  fprintf(ftbl,
            "Error: LINK must either be 2 16 bit words per frame %s",
	    "(bits_per_word=0,words_per_frame=1)\n              or be 1 %s",
            "32 bit word per frame (bits_per_word=1,words_per_frame=0)\n");
	}

	if(gen->nibble == 0)  {
	  fprintf(stderr, "Need nibble=1 for output_mode=2\n");
	  fprintf(ftbl, "Need nibble=1 for output_mode=2\n");
	}

/* Gary says this isn't needed anymore 
	if(gen->link && (gen->ckdiv ==0)) fprintf(stderr, 
	    "Error: ckdiv must be > 0 in LINK mode\n");
*/
	if(gen->rev && (gen->nibble || !packed)) {
	  fprintf(stderr, 
	"Warning: You probably do not want reverse_iq=1\n");
	  fprintf(ftbl, 
	"Warning: You probably do not want reverse_iq=1\n");
	}
	if((gen->rev==0) && packed && gen->link) {
	  fprintf(stderr, 
	"Warning: This is a packed LINK so you probably want reverse_iq=1\n");
	  fprintf(ftbl, 
	"Warning: This is a packed LINK so you probably want reverse_iq=1\n");
	}
	if((gen->ckdiv+1)*osamplerate*4*(gen->bnperword+1) > 1.0) {
	fprintf(stderr, "Error: Output rate is too high\n");
	fprintf(ftbl, "Error: Output rate is too high\n");
	}
	if((gen->enrdy==1) && gen->link) {
	fprintf(stderr, "Error: This is a LINK so you need enrdy=0\n");
	fprintf(ftbl, "Error: This is a LINK so you need enrdy=0\n");
	}
	if((gen->ensa!=1)|| (gen->ensb!=1)|| (gen->ensc!=1)|| (gen->ensd!=1)) {
	fprintf(stderr, "Error: Need to enable enp0-3 to get your output\n");
	fprintf(ftbl, "Error: Need to enable enp0-3 to get your output\n");
	}
	if((gen->enpar==1)) {
	  fprintf(stderr, 
	"Warning: enpar not needed in nibble mode. It just adds noise\n");
	  fprintf(ftbl, 
	"Warning: enpar not needed in nibble mode. It just adds noise\n");
	}
	if(realout && gen->rev && (gen->blocklength==0)) {
	  fprintf(stderr, 
	"Error: Can't reverse_iq with real_only and block_size=0\n");
	  fprintf(ftbl, 
	"Error: Can't reverse_iq with real_only and block_size=0\n");
	}
	if((gen->fl<=gen->wordsperframe) && (gen->bnperword==0)) {
	  fprintf(stderr,
	 "Error: frame_length must be larger than words_per_frame for %s\n",
	 "nibble mode with 16 bit words");
	  fprintf(ftbl,
	 "Error: frame_length must be larger than words_per_frame for %s\n",
	 "nibble mode with 16 bit words");
	}
	  


    } else if (gen->os == PARALLEL) { /* parallel output checking */
	if((gen->ckdiv+1)*osamplerate >= 1.0)  {
	fprintf(stderr, "Error: Output rate is too high\n");
	fprintf(ftbl, "Error: Output rate is too high\n");
	}
	if(gen->parallel==0)  {
	fprintf(stderr, "Error: Parallel must be enabled with output_mode=3\n");
	fprintf(ftbl, "Error: Parallel must be enabled with output_mode=3\n");
	}
        if(gen->no != 0) {
	  fprintf(stderr, 
		"Error: nserial must be 0 in parallel modes\n");
	  fprintf(ftbl, 
		"Error: nserial must be 0 in parallel modes\n");
	}
	if((gen->ensa+gen->ensb+gen->ensc+gen->ensd) != 4) {
	  fprintf(stderr,
	  "Error: enp0-3 must be high in parallel mode\n");
	  fprintf(ftbl,
	  "Error: enp0-3 must be high in parallel mode\n");
	}
	if(gen->enpar==0) {
	  fprintf(stderr,
	  "Error: enpar must be high in parallel mode\n");
	  fprintf(ftbl,
	  "Error: enpar must be high in parallel mode\n");
	}
	if(gen->encko==0) {
	  fprintf(stderr,
	  "Error: ensck must be high in parallel mode\n");
	  fprintf(ftbl,
	  "Error: ensck must be high in parallel mode\n");
	}
	if(gen->enrdy==0) {
	  fprintf(stderr,
	  "Error: enrdy must be high in parallel mode\n");
	  fprintf(ftbl,
	  "Error: enrdy must be high in parallel mode\n");
	}
	if(gen->ensfs==0) {
	  fprintf(stderr,
	  "Error: ensfs must be high in parallel mode\n");
	  fprintf(ftbl,
	  "Error: ensfs must be high in parallel mode\n");
	}
        if(gen->link != 0)  {
		fprintf(stderr, "Error: link must be zero in parallel mode\n");
		fprintf(ftbl, "Error: link must be zero in parallel mode\n");
	}
        if(gen->nibble != 0)  {
		fprintf(stderr, "Error: nibble must be zero in parallel mode\n");
		fprintf(ftbl, "Error: nibble must be zero in parallel mode\n");
	}
    }
}
